# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import json
import uuid

from tornado.escape import json_decode, utf8

from jupyter_server.gateway.connections import GatewayWebSocketConnection
from jupyter_server.gateway.managers import GatewayKernelManager
from jupyter_server.services.kernels.connection.base import BaseKernelWebsocketConnection
from jupyter_server.services.kernels.connection.channels import ZMQChannelsWebsocketConnection

class StartingReportingWebsocketConnection(GatewayWebSocketConnection):
    """Extension of GatewayWebSocketConnection that reports a starting status on connection.

    The purpose of this class is to bridge the time period between when the websocket
    connection from the client is created and the websocket connection to the Gateway
    server is created.

    During that time, the JupyterLab UI believes that it is connected to the running
    kernel, but it has not yet received any status messages from the kernel, so it will
    display a kernel status of "Unknown".

    That "Unknown" status is not very helpful to users because they have no idea why
    the kernel status is unknown and have no indication that something is still
    happening under the hood.

    To improve that, we report a provisional status as soon as the client connection
    is established. This provisional status will be replaced by the real status
    reported by the kernel as soon as the backend kernel connection is established.

    The only kernel statuses supported by the JupyterLab UI are "starting", "idle",
    "busy", "restarting", and "dead".

    Of those, the "starting" message is the closest match to what is going on, so
    we use that one.

    However, the "starting" message is only supposed to be reported once, so we
    also intercept any "starting" messages received from the kernel and discard
    them, as we know we will have already reported this status.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def connect(self):
        # The kernel message format is defined
        # [here](https://jupyter-client.readthedocs.io/en/latest/messaging.html#general-message-format).
        status_message_id = str(uuid.uuid4())
        status_message = {
            "header": {
                "msg_id": status_message_id,
                "session": self.kernel_id,
                "username": "username",
                "date": datetime.datetime.utcnow().isoformat(),
                "msg_type": "status",
                "version": "5.3",
            },
            "parent_header": {},
            "metadata": {},
            "msg_id": status_message_id,
            "msg_type": "status",
            "channel": "iopub",
            "content": {
                "execution_state": "starting",
            },
            "buffers": [],
        }
        super().handle_outgoing_message(json.dumps(status_message))
        return await super().connect()

    def is_starting_message(self, incoming_msg):
        try:
            msg = json_decode(utf8(incoming_msg))
            if msg.get("content", {}).get("execution_state", "") == "starting":
                return True
        except Exception as ex:
            pass
        return False

    def handle_outgoing_message(self, incoming_msg, *args, **kwargs):
        if self.is_starting_message(incoming_msg):
            # We already sent a starting message, so drop this one.
            return
        return super().handle_outgoing_message(incoming_msg, *args, **kwargs)


class DelegatingWebsocketConnection(BaseKernelWebsocketConnection):
    """Implementation of BaseKernelWebsocketConnection that delegates to another connection.

    If the parent KernelManager instance is for a remote kernel (i.e. it is a
    GatewayKernelManager), then the delegate is an instance of
    StartingReportingWebsocketConnection, which extends GatewayWebSocketConnection.

    Otherwise, it is an instance of ZMQChannelsWebsocketConnection.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        delegate_class = ZMQChannelsWebsocketConnection
        if self.kernel_manager.is_remote:
            delegate_class = StartingReportingWebsocketConnection
        self.delegate = delegate_class(
            parent=self.kernel_manager.delegate,
            websocket_handler=self.websocket_handler,
            config=self.config)

    async def connect(self):
        return await self.delegate.connect()

    async def disconnect(self):
        return await self.delegate.disconnect()

    def handle_incoming_message(self, msg):
        return self.delegate.handle_incoming_message(msg)

    def handle_outgoing_message(self, stream, msg):
        return self.delegate.handle_outgoing_message(stream, msg)

    # Prepare actually comes from ZMQChannelsWebsocketConnection.
    #
    # It is called by the jupyter_server kernels websocket handler if present, so
    # we provide an implemention of it in case the delegate is an instance of the
    # ZMQChannelWebsocketConnection class.
    async def prepare(self):
        if hasattr(self.delegate, "prepare"):
            return await self.delegate.prepare()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from google.cloud.jupyter_config import configure_gateway_client

from kernels_mixer.kernelspecs import MixingKernelSpecManager
from kernels_mixer.kernels import MixingMappingKernelManager
from kernels_mixer.websockets import DelegatingWebsocketConnection

from jupyter_server.services.sessions.sessionmanager import SessionManager

def configure_kernels_mixer(c):
  """Helper method for configuring the given Config object to use the GCP kernel gateway."""
  configure_gateway_client(c)
  c.ServerApp.kernel_spec_manager_class = MixingKernelSpecManager
  c.ServerApp.kernel_manager_class = MixingMappingKernelManager
  c.ServerApp.session_manager_class = SessionManager
  c.ServerApp.kernel_websocket_connection_class = DelegatingWebsocketConnection



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jupyter_client.kernelspec import KernelSpecManager
from jupyter_core.utils import run_sync
from jupyter_server.gateway.managers import GatewayKernelSpecManager

from traitlets import Type, Unicode, default


def append_display_name(spec, suffix):
    """Append the given suffix onto the display name of the given kernelspec.

    The supplied kernelspec is updated in place.

    Args:
      spec: Either an object with a "display_name" attribute, or a
            dictionary with a "display_name" string field.
      suffix: A string suffix to append to the spec's display name.
    """
    if hasattr(spec, "display_name"):
        spec.display_name = spec.display_name + suffix
    else:
        spec["display_name"] = spec.get("display_name", "") + suffix


class MixingKernelSpecManager(KernelSpecManager):

    local_display_name_suffix = Unicode(
        " (Local)",
        config=True,
        help="Suffix added to the display names of local kernels.",
    )

    remote_display_name_suffix = Unicode(
        " (Remote)",
        config=True,
        help="Suffix added to the display names of remote kernels.",
    )

    local_kernel_spec_manager_class = Type(
        config=True,
        default_value=KernelSpecManager,
        help="""
        The kernel spec manager class to use for local kernels.

        Must be a subclass of `jupyter_client.kernelspec.KernelSpecManager`.""",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.local_manager = self.local_kernel_spec_manager_class(*args, **kwargs)
        self.remote_manager= GatewayKernelSpecManager(*args, **kwargs)
        self._local_kernels = set()
        self._remote_kernels = set()

    def is_remote(self, kernel_name):
        return kernel_name in self._remote_kernels

    def get_all_specs(self):
        """Get a list of all kernelspecs supported.

        This is a combination of the kernelspecs supported by both the local and remote
        kernel spec managers.

        In case a kernel name is supported by both the local and remote kernel spec
        manager, the local kernel spec manager's version is used and the remote
        one is ignored.

        The return value is a dictionary mapping kernel names to kernelspecs.

        Each kernelspec is a dictionary with the following keys:
          1. "name": The name of the kernel, which must adhere to the Jupyter API naming rules.
          2. "spec": A "KernelSpecFile" resource, which itself is a dictionary.
          3. "resources": A dictionary mapping resource names to URIs.

        A KernelSpecFile dictionary is described here:
          https://github.com/jupyter-server/jupyter_server/blob/c5dc0f696f376e1db5a9a0cbcebb40a0bf98875c/jupyter_server/services/api/api.yaml#L781

        Of particular note, it contains one entry with a key of "display_name" whose value is
        the name for the kernel displayed in the JupyterLab launcher and kernel picker.

        Returns:
          A map from kernel names (str) to kernelspecs.
        """
        ks = self.local_manager.get_all_specs()
        for name, kernelspec in ks.items():
            spec = kernelspec.get("spec", {})
            append_display_name(spec, self.local_display_name_suffix)
            self._local_kernels = self._local_kernels | {name}
        try:
            remote_ks = run_sync(self.remote_manager.get_all_specs)()
            for name, kernelspec in remote_ks.items():
                if name not in self._local_kernels:
                    spec = kernelspec.get("spec", {})
                    append_display_name(spec, self.remote_display_name_suffix)
                    ks[name] = kernelspec
                    self._remote_kernels = self._remote_kernels | {name}
                    
        except Exception as ex:
            self.log.exception('Failure listing remote kernelspecs: %s', ex)
            # Otherwise ignore the exception, so that local kernels are still usable.
        self.log.debug(f'Found {len(self._local_kernels)} local kernels: {self._local_kernels}')
        self.log.debug(f'Found {len(self._remote_kernels)} remote kernels: {self._remote_kernels}')
        return ks

    def get_original_kernel_spec(self, kernel_name, *args, **kwargs):
        if self.is_remote(kernel_name):
            self.log.debug(f'Looking up remote kernel {kernel_name}...')
            return run_sync(self.remote_manager.get_kernel_spec)(kernel_name, *args, **kwargs)
        self.log.debug(f'Looking up local kernel {kernel_name}...')
        return self.local_manager.get_kernel_spec(kernel_name, *args, **kwargs)

    def get_kernel_spec(self, kernel_name, *args, **kwargs):
        spec = self.get_original_kernel_spec(kernel_name, *args, **kwargs)
        suffix = self.local_display_name_suffix
        if self.is_remote(kernel_name):
            suffix = self.remote_display_name_suffix
        append_display_name(spec, suffix)
        return spec

    async def get_kernel_spec_resource(self, kernel_name, path):
        if self.is_remote(kernel_name):
            self.log.debug(f'Looking up remote kernel spec resource for {kernel_name}...')
            return await self.remote_manager.get_kernel_spec_resource(kernel_name, path)
        return None
